package main

import (
	"encoding/json"
	"errors"
	"fmt"
	"log"
	"net/http"
	"net/url"
	"strings"

	"github.com/go-stomp/stomp"
	"github.com/gorilla/websocket"
)

const (
	// 团队令牌
	token = "EDD6-8D31-C838-8C3F"
	// 设备UUID
	uuid  = "3DCD-1188-9235-3594"

	// Janus地址
	baseurl = "http://127.0.0.1:8080/api/janus/"
)

type Response struct {
	// Janus返回数据格式
	Code    int               `json:"code"`
	Message string            `json:"message"`
	Data    map[string]string `json:"data"`
}

var (
	// Janus返回的错误
	errInvalidToken    = errors.New("invalid token")
	errDeviceNotExist  = errors.New("device does not exist")
	errUnknownResponse = errors.New("unknown response")
)

func main() {
	// 向Janus获取WebSocket地址以及订阅Topic
	wsPath, topic, err := getWebsocketAddress(token, uuid)
	if err != nil {
		log.Fatal(err)
	}

	// 连接WebSocket并使用STOMP协议通讯
	client, err := getSTOMPClient(wsPath)
	if err != nil {
		log.Fatal(err)
	}

	// 订阅Topic
	sub, err := client.Subscribe(topic, stomp.AckAuto, nil)
	if err != nil {
		log.Fatal(err)
	}

	// 订阅存在时
	for sub.Active() {
		// 读取订阅内容
		msg, err := sub.Read()
		if err != nil {
			log.Fatal(err)
		}

		// 打印订阅内容
		fmt.Printf("Received message: %s from topic %s\n", msg.Body, msg.Destination)
	}
}

// getWebsocketAddress is a function that sends a request to Janus with token and uuid specified, parses the response
// and return the corresponding WebSocket path and STOMP topic, or error if returned.
func getWebsocketAddress(token, uuid string) (path, topic string, err error) {
	u, err := url.Parse(baseurl + "websocket")
	if err != nil {
		return "", "", err
	}

	q := u.Query()
	q.Set("token", token)
	q.Set("uuid", uuid)

	u.RawQuery = q.Encode()

	resp, err := http.DefaultClient.Get(u.String())
	if err != nil {
		return "", "", err
	}
	defer resp.Body.Close()

	var data Response
	dec := json.NewDecoder(resp.Body)
	if err := dec.Decode(&data); err != nil {
		return "", "", err
	}

	switch data.Code {
	case 200:
		return data.Data["websocket"], data.Data["topic"], nil
	case 401:
		return "", "", errInvalidToken
	case 1003:
		return "", "", errDeviceNotExist
	default:
		return "", "", errUnknownResponse
	}
}

// getSTOMPClient initiates a WebSocket connection to the URL specified and do a STOMP protocol handshake, then returns
// the connection for use
func getSTOMPClient(fullUrl string) (*stomp.Conn, error) {
	u, err := url.ParseRequestURI(fullUrl)
	if err != nil {
		return nil, err
	}

	var urlStr string
	urlStr = strings.Replace(u.String(), "http", "ws", 1)

	conn, _, err := websocket.DefaultDialer.Dial(urlStr, nil)
	if err != nil {
		return nil, err
	}

	return stomp.Connect(&wsWrapper{conn})
}

// wsWrapper is a wrapper object that implements io.ReadWriteCloser for use with STOMP client
type wsWrapper struct {
	conn *websocket.Conn
}

func (w *wsWrapper) Read(p []byte) (n int, err error) {
	_, b, err := w.conn.ReadMessage()
	return copy(p, b), err
}

func (w *wsWrapper) Write(p []byte) (n int, err error) {
	return len(p), w.conn.WriteMessage(websocket.TextMessage, p)
}

func (w *wsWrapper) Close() error {
	return w.conn.Close()
}
